/**************************************************************
stfp.c - bc635PCI Solaris 8 - 10 Device Driver

PROJECT: bc635PCI Time and Frequency Processor

V1 DEVELOPMENT SYSTEM: Sun SPARCstation1/Solaris 2.5.1
V2 DEVELOPMENT SYSTEM: Sun Ultra AX PCI uSPARC motherboard. (EIS Fusion-1)

TAB STOPS: 4

Copyright (c) 1994-2009 Symmetricom, Inc.
              San Jose, CA. All rights reserved.

modification history
====================
V7.0  21 Mar 08, jsy  Add functions to support V2 hardware
V4.4  06 Nov 00, sd   Added new DPRAM functions for 12083 H/W
V3.1  15 Mar 00, sd   Fixed the Strobe function
V2.3  21 Jun 99, ram  Ported to 32/64-bit Solaris 7
V2.2  08 May 98, crs  Optimized Get method to use union, add delay to fmt chg.
V2.1, 02 May 98, dvb  Added delay to mode chg., corrected send-time for PCI
V2.0, 23 Mar 98, ram  Ported sbus version to PCI and 2.5.1 Solaris
V1.0, 18 Jan 95, mlt  version complete
V1.0, 19 Nov 94, jwf  begin porting from SunOS 4.1.1 version
**************************************************************/

/*--------------------------  INCLUDE FILES  --------------------------------*/

#include <sys/types.h>
/*
#include <sys/dditypes.h>
#include <sys/ksynch.h>
*/

#include <sys/param.h>
#include <sys/errno.h>
#include <sys/signal.h>
#include <sys/uio.h>
#include <sys/buf.h>
#undef major
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/kmem.h>
#include <sys/poll.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/stat.h>
#include <sys/procset.h>

#include <sys/ddi.h>
#include <sys/sunddi.h>

/*
#include <sys/proc.h>
*/

#include <sys/pci.h>

#include "stfpio.h"

/*----------------------------  MACROS  -------------------------------------*/

#define STFP_DP_GETB(p)                 (ddi_get8(psu->Dprhandle, (p)))
#define STFP_DP_PUTB(p, val)            \
(ddi_put8(psu->Dprhandle, (p), (unsigned char)((val) & 0xff)))

#define STFP_REG_GET32(p)               (ddi_get32(psu->Regshandle, (p)))
#define STFP_REG_PUT32(p, val)          (ddi_put32(psu->Regshandle, (uint32_t *)(p), (val)))

#define STFP_REG_OREQUAL(p, val)        (STFP_REG_PUT32((p), STFP_REG_GET32(p)|(val)))
#define STFP_REG_ANDEQUAL(p, val)       (STFP_REG_PUT32((p), STFP_REG_GET32(p)&(val)))


#define STFP_INTR_PENDING               (STFP_REG_GET32(&psu->pRegs->mask) & \
                                         STFP_REG_GET32(&psu->pRegs->intstat) & \
                                         INT_BITMASK)

#define TOP_DPRAM 0x800

/*------------------------  DEBUGGING SWITCHES  -----------------------------*/

int32_t stfp_debug = 1;
#define DEBUG1 if (stfp_debug > 0) cmn_err
#define DEBUG2 if (stfp_debug > 1) cmn_err
#define DEBUG3 if (stfp_debug > 2) cmn_err

#undef DUMP_MAPPING

/*-------------------------  DATA  STRUCTURES  ------------------------------*/

struct stfp_regs {          /* hardware registers */
    uint32_t timereq;       /* time request (for time[] regs) */
    uint32_t eventreq;      /* time request (for event[] regs) */
    uint32_t unlock;        /* time capture lockout release */
    uint32_t unlock2;
    uint32_t control;       /* misc. control register */
    uint32_t ack;           /* acknowledge register */
    uint32_t mask;          /* interrupt mask register */
    uint32_t intstat;       /* interrupt status register */
    uint32_t minstrb;       /* minor strobe time */
    uint32_t mijstrb;       /* major strobe time */
    uint32_t event2[2];
    uint32_t time[2];       /* minor/major time hold */
    uint32_t event[2];      /* minor/major event */
    uint32_t reserved;
    uint32_t unlock3;
    uint32_t event3[2];
};

struct stfp_unit {
    dev_info_t              *dip;           /* device info */
    struct stfp_regs        *pRegs;         /* Device control regs */
    u_char                  *pDpr;          /* DPRAM base pointer */
    u_char                  *pIn;           /* usr input area in DPRAM */
    u_char                  *pOut;          /* output area (to user) in DPRAM */
    u_char                  *pGps;          /* GPS packet area in DPRAM */
    u_char                  *pYear;         /* year area in DPRAM */
    uint32_t                dpramSize;      /* size of DPRAM */
    int32_t                 intServiced;    /* last interrupt(s) seviced */
    int32_t                 intMaskSig;     /* Interrupt mask for signals */
    void                    *pref;          /* process reference for signals */
    kmutex_t                mutex;          /* mutex */
    kcondvar_t              cv;             /* cond variavle (sleep/wakeup) */
    ddi_iblock_cookie_t     iblock_cookie;  /* interrupt block cookie */
    ddi_idevice_cookie_t    idevice_cookie; /* device cookie */
    ddi_acc_handle_t        Dprhandle;
    ddi_acc_handle_t        Regshandle;
    u_char                  revID;          /* hardware revision */
};

/*---------------------  DRIVER  DECLARATIONS  ------------------------------*/

static int stfp_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int stfp_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static int stfp_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
                        void **result);
static int stfp_open(dev_t *devp, int flag, int otyp, cred_t *credp);
static int stfp_close(dev_t dev, int flag, int otyp, cred_t *credp);
static int stfp_read(dev_t dev, struct uio *uiop, cred_t *credp);
static int stfp_write(dev_t dev, struct uio *uiop, cred_t *credp);
static int stfp_ioctl(dev_t dev, int cmd, intptr_t arg, int flag, 
                      cred_t *credp, int *rvalp);
static u_int stfp_intr(caddr_t arg);

/*------------------ DATA STRUCTURES ----------------------------------------*/

static void *stfpstatep;        /* Ptr to instance state struct */

/*   BOOT STATES
 * 0 - Attach not run: Most likely card not present or PCI interface is dead.
 * 1 - Card dualport RAM offsets did not get initalized.  HC11 is dead?
 * 2 - Card booted correctly.
 */
int boot_state = 0;

static struct cb_ops stfpcb_ops = {
    stfp_open,              /* open */
    stfp_close,             /* close */
    nodev,                  /* strategy */
    nodev,                  /* print */
    nodev,                  /* dump */
    stfp_read,              /* read */
    stfp_write,             /* write */
    stfp_ioctl,             /* ioctl */
    nodev,                  /* devmap */
    nodev,                  /* mmap */
    nodev,                  /* segmap */
    nochpoll,               /* poll */
    ddi_prop_op,            /* cb_prop_op */
    0,                      /* streamtab  */
    D_NEW | D_MTSAFE        /* Driver comaptibility flag */
};
 
/* Device driver ops vector */
struct dev_ops stfp_dev_ops = {
    DEVO_REV,               /* devo_rev */
    0,                      /* refcnt  */
    stfp_getinfo,           /* info */
    nulldev,                /* identify */
    nulldev,                /* probe */
    stfp_attach,            /* attach */
    stfp_detach,            /* detach */
    nodev,                  /* reset */
    &stfpcb_ops,            /* driver operations */
    0,                      /* bus operations */
};
 
/*------------------ MODULE LINKAGE -----------------------------------------*/

static struct modldrv modldrv = 
{
    &mod_driverops,                                           /* Adrs of std driver ops.*/
    "Symmetricom bc635/637PCIe-V2 Time and Frequency Processor",  /* Returned by modinfo cmd.*/
    &stfp_dev_ops,                                            /* This driver's ops.*/
};

static struct modlinkage modlinkage = 
{
    MODREV_1,               /* Loadable modules system rev */
    &modldrv,               /* Null terminated list of */
    NULL                    /* linkage structures. */
};

/*--------------------------- LOCAL FUNCTIONS -------------------------------*/

void rdTimeTv(struct stfp_unit *psu, uint32_t *pTmReg, struct stfp_timeval *pTmT);
void rdTime(struct stfp_unit *psu, uint32_t *pTmReg, u_char *pY, struct stfp_time *pTmS);
int sendCommand(struct stfp_unit *psu);

/*--------------------- DRIVER ENTRY POINTS ---------------------------------*/

/*****************************************************************************
* _init - initialize the loadable driver module
*/
int
_init(void)
{
    int error;      /* Return value holder.*/

    DEBUG3(CE_CONT, "_init: Called!\n");

    if((error = ddi_soft_state_init(&stfpstatep,    /* Init state ptr and */
        sizeof(struct stfp_unit), 1)) != 0)         /* struct size with sys.*/
    {
        DEBUG1(CE_CONT, "_init: Failed ddi_soft_state!\n");
        return(error);
    }

    if((error = mod_install(&modlinkage)) != 0)     /* Install driver.*/
    {
        DEBUG1(CE_CONT, "_init: Failed mod_install!\n");
        ddi_soft_state_fini(&stfpstatep);           /* Devinstall on error.*/
    }

    return(error);  /* Return val from mod_install*/
}

/*****************************************************************************
* _info - returns the loadable driver module information
*/
int
_info(struct modinfo *modinfop)
{
    DEBUG3(CE_CONT, "_info: Called!\n");

    return(mod_info(&modlinkage, modinfop));    /* Load driver.*/
}
 
/*****************************************************************************
* _fini - prepares loadable driver module for unloading
*/
int
_fini( void )
{
    int status;     /* Returned status holder.*/

    DEBUG3(CE_CONT, "_fini: Called!\n");

    if (boot_state == 0)
    {
        cmn_err(CE_CONT, "635PCI not found-Try sys reboot with power cycle (-r)\n");
    }

    if((status = mod_remove(&modlinkage)) != 0)     /* Unload driver.*/
    {
        DEBUG1(CE_CONT, "_fini: Error on mod_remove: %d!\n", status);
        goto exit;
    }
 
    /* Release, here, per-module resources as required!! */
    ddi_soft_state_fini(&stfpstatep);      /* What does this do?? */

exit:
    return(status);
}

#ifdef DUMP_DEBUG

void
dump_pci_config(dev_info_t *dip, int nbytes)
{
    ddi_acc_handle_t handle;
    uint32_t cl;
    uint32_t offset;
    uint32_t base0;
    uint32_t size0;

    if (pci_config_setup(dip, &handle) != DDI_SUCCESS)
    {
        DEBUG1(CE_CONT, "dump_pci_config: pci_config_setup failed!");
        return;
    }

    DEBUG1(CE_CONT, "Configuration Space:\n");
    for (offset=0;offset<nbytes;offset+=4)
    {
        cl = pci_config_get32(handle, offset);
        DEBUG1(CE_CONT, "0x%x ", cl);
        if ((offset%0x10)==0xC)
            DEBUG1(CE_CONT, "\n");
    }
    DEBUG1(CE_CONT, "\n");

    pci_config_teardown(&handle);
}

void
dump_dualport_ram(ddi_acc_handle_t *handlep, uchar_t *base, uint32_t base_offset, uint32_t length)
{
    uint32_t offset;
    unsigned char valb;

    DEBUG1(CE_CONT, "0x%x of Dual-Port RAM:	", base_offset);

    for (offset=0; offset<length; offset++)
    {
        valb = ddi_get8(*handlep, base+base_offset+offset);
        DEBUG1(CE_CONT, " 0x%x", valb);
    }

    DEBUG1(CE_CONT, "\n");
}

void
dump_registers(ddi_acc_handle_t *handlep, uchar_t base, uint32_t base_offset, uint32_t length)
{
    uint32_t offset;
    uint32_t vall;

    DEBUG1(CE_CONT, "0x%x of Registers:	", base_offset);

    length = (((length-1)/4)+1)*4;
    for (offset=0; offset<length; offset+=4)
    {
        vall = ddi_get32(*handlep, base+base_offset+offset);
        DEBUG1(CE_CONT, " 0x%X", vall);
    }

    DEBUG1(CE_CONT, "\n");
}

void
hc11_diagnostic(struct stfp_unit *psu)
{
    int i;
    unsigned char var1, var2, var3, var4;

    /* clear ack command done bit */
    STFP_REG_PUT32(&psu->pRegs->ack, ACK_COMMAND);
    /* force bc635PCI to take action */
    STFP_REG_PUT32(&psu->pRegs->ack, ACK_ACTION);

    for (i=0;((STFP_REG_GET32(&psu->pRegs->ack) & ACK_COMMAND) == 0) && i<20;i++)
        delay(drv_usectohz(1<<i));

    if(i >= 20) {   /* check if timeout occured */
        DEBUG1(CE_CONT, " HC11 is stopped!\n");
    }
    else {
        DEBUG3(CE_CONT, " HC11 is responding!\n");
    }
}
#endif

/*****************************************************************************
* stfp_attach - perform device specific initialization
*/

static int
stfp_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
    int instance;
    struct stfp_unit *psu;
    char namesfx[3];        /* used to create name of the device entry */
    dev_t dev;
    ddi_device_acc_attr_t accattr;
    ddi_acc_handle_t handle;
    uint32_t ui_val;
    uint_t bar_reg, bar_dp;
#ifdef DUMP_MAPPING
    int nmap, i;
    off_t szReg;
#endif

    DEBUG3(CE_CONT, "Entering stfp_attach:\n");
	
    if (cmd != DDI_ATTACH)
        return (DDI_FAILURE);

    instance = ddi_get_instance(dip);
    DEBUG3(CE_CONT, "stfp_attach: instance = %d\n", instance);
	
    if (ddi_soft_state_zalloc(stfpstatep, instance) != 0)
    {
        DEBUG1(CE_CONT, "Can't zalloc state struct: %s %d\n",
               ddi_get_name(dip), instance);
        return (DDI_FAILURE);
    }
	
    psu = ddi_get_soft_state (stfpstatep, instance);
	
    /* initialize stfp_unit stuff here */
    psu->dip = dip;     /* save info handle for later */
    psu->pDpr = NULL;
    psu->pRegs = NULL;
	
    /* register interrupts here */
    if(ddi_intr_hilevel(dip, 0) != 0)
    {
        /* High level interrupts are not supported */
        DEBUG3(CE_CONT, "High level interrupts are not supported: %s\n",
               ddi_get_name(dip));
        goto fail1;
    }
    else
    {
        /* Add low level interrupt handler and initialize
         * corresponding mutual execution lock
         */
        if(ddi_get_iblock_cookie(dip, 0, &psu->iblock_cookie) != DDI_SUCCESS)
        {
            DEBUG1(CE_CONT, "ddi_attach:  ddi_get_iblock_cookie failed!\n");
            goto fail1;
        }

        mutex_init(&psu->mutex, "stfp mutex", MUTEX_DRIVER,
                   (void *)psu->iblock_cookie);

        if(ddi_add_intr(dip, 0, &psu->iblock_cookie, &psu->idevice_cookie,
                stfp_intr, (caddr_t)psu) != DDI_SUCCESS)
        {
            DEBUG1(CE_CONT, "ddi_attach:  ddi_add_intr failed!\n");
            goto fail2;
        }
    }
	
    /* initialize mutual condition variable */
    cv_init(&psu->cv, "stfp cv", CV_DRIVER, NULL);

#ifdef DUMP_MAPPING
    if (ddi_dev_nregs(dip, &nmap) != DDI_SUCCESS)
    {
        DEBUG1(CE_CONT, "ddi_dev_nregs failed\n");
        goto fail3;
    }
    for (i = 0; i < nmap; i++)
    {
        if (ddi_dev_regsize(dip, i, &szReg) != DDI_SUCCESS)
        {
            DEBUG1(CE_CONT, "ddi_dev_regsize on %d map failed\n", i);
        }
        else
        {
            DEBUG1(CE_CONT, "reg map %d size is %d\n", i, szReg);
        }
    }
#endif

    // Retrieve hardware revision
    if (pci_config_setup(dip, &handle) != DDI_SUCCESS)
    {
        DEBUG1(CE_CONT, "ddi_attach: pci_config_setup failed!");
        goto fail3;
    }

    // Revision ID is at 0x8
    ui_val = pci_config_get32(handle, 0x8);
    psu->revID = (u_char)(ui_val & 0xFF);

    pci_config_teardown(&handle);

    if (psu->revID >= BC_PCI_V2_REV_ID_START &&
        psu->revID <= BC_PCI_V2_REV_ID_END)
    {
        bar_reg = 2;
        bar_dp  = 3;
    }
    else
    {
        bar_reg = 1;
        bar_dp  = 2;
    }

    accattr.devacc_attr_version = DDI_DEVICE_ATTR_V0;
    accattr.devacc_attr_endian_flags = DDI_STRUCTURE_BE_ACC;
    accattr.devacc_attr_dataorder = DDI_STRICTORDER_ACC;

    if (ddi_regs_map_setup(dip, bar_dp, (caddr_t *)&psu->pDpr,
        (offset_t)0, (offset_t)0, &accattr, &psu->Dprhandle) != DDI_SUCCESS)
    {
        DEBUG1(CE_CONT, "ddi_regs_map_setup failed\n");
        goto fail3;
    }

    accattr.devacc_attr_version = DDI_DEVICE_ATTR_V0;
    accattr.devacc_attr_endian_flags = DDI_STRUCTURE_LE_ACC;
    accattr.devacc_attr_dataorder = DDI_STRICTORDER_ACC;

    if (ddi_regs_map_setup(dip, bar_reg, (caddr_t *)&psu->pRegs,
        (offset_t)0, (offset_t)0, &accattr, &psu->Regshandle) != DDI_SUCCESS)
    {
        DEBUG1(CE_CONT, "ddi_regs_map_setup failed\n");
        goto fail4;
    }

    DEBUG3(CE_CONT, "ddi_regs_map_setup - pDpr = 0x%x, pRegs = 0x%x\n", psu->pDpr, psu->pRegs);
    DEBUG3(CE_CONT, "\n");

    psu->pIn = psu->pDpr + ((STFP_DP_GETB(psu->pDpr+TOP_DPRAM - 1))
                + (STFP_DP_GETB(psu->pDpr+TOP_DPRAM - 2) << 8));
    psu->pOut = psu->pDpr + ((STFP_DP_GETB( psu->pDpr+TOP_DPRAM - 3))
                + (STFP_DP_GETB(psu->pDpr+TOP_DPRAM - 4) << 8));
    psu->pGps = psu->pDpr + ((STFP_DP_GETB(psu->pDpr+TOP_DPRAM - 5))
                + (STFP_DP_GETB(psu->pDpr+TOP_DPRAM - 6) << 8));
    psu->pYear = psu->pDpr + ((STFP_DP_GETB(psu->pDpr+TOP_DPRAM - 7))
                + (STFP_DP_GETB(psu->pDpr+TOP_DPRAM - 8) << 8));

    DEBUG3(CE_CONT, "pDpr = %x\n", psu->pDpr);
    DEBUG3(CE_CONT, "pIn = %x\n", psu->pIn);
    DEBUG3(CE_CONT, "pOut = %x\n", psu->pOut);
    DEBUG3(CE_CONT, "pGps = %x\n", psu->pGps);
    DEBUG3(CE_CONT, "pYear = %x\n", psu->pYear);

    if ((psu->pIn - psu->pDpr) == 0x102)
    {
        cmn_err(CE_CONT, "Symmetricom bc635/637PCIe-V2 Time and Frequency Processor\n");
        cmn_err(CE_CONT, "Copyright (c) 1998-2009 Symmetricom, Inc.  all rights reserved\n");
        boot_state = 2;
    }
    else
    {
        boot_state = 1;
        cmn_err(CE_CONT, "bc635/637PCIe-V2 Did Not Boot Correctly - Try system warm reboot\n");
        /* HC11 diagnostic */
        goto fail5;
    }

    sprintf(namesfx, "%d", instance);
    DEBUG3(CE_CONT, "ddi_create_minor_node(%s)\n", namesfx);

    if (ddi_create_minor_node (dip, namesfx, S_IFCHR, instance,
        DDI_PSEUDO, (int)NULL) != DDI_SUCCESS)
    {
        DEBUG1(CE_CONT, "Can't create minor node: %s %d\n",
               ddi_get_name(dip), instance);
        goto fail5;
    }
	
    ddi_report_dev (dip);
    return (DDI_SUCCESS);

fail5:
    ddi_regs_map_free(&psu->Dprhandle);

fail4:
    ddi_regs_map_free(&psu->Regshandle);

fail3:
    cv_destroy(&psu->cv);
    ddi_remove_intr(dip, 0, psu->iblock_cookie);

fail2:
    mutex_destroy(&psu->mutex);

fail1:
    ddi_soft_state_free(stfpstatep, instance);
    return (DDI_FAILURE);
}

/*****************************************************************************
* stfp_detach - remove device specific state
*/

static int
stfp_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
    struct stfp_unit *psu;
    int instance;
	
    DEBUG3(CE_CONT, "Entering stfp_detach:\n");
	
    if (cmd != DDI_DETACH)
        return (DDI_FAILURE);
	
    instance = ddi_get_instance(dip);

    /* remove minor node */
    psu = ddi_get_soft_state (stfpstatep, instance);
    ddi_remove_minor_node(dip, NULL);
	
    /* remove interrupts here */
    STFP_REG_PUT32(&psu->pRegs->mask, 0);
    ddi_remove_intr(dip, 0, psu->iblock_cookie);

    mutex_destroy(&psu->mutex);

    /* destroy mutual condition variable */
    cv_destroy(&psu->cv);

    /* release resources and unmap registers */
    ddi_regs_map_free(&psu->Dprhandle);
    ddi_regs_map_free(&psu->Regshandle);

    ddi_soft_state_free(stfpstatep, instance);

    return (DDI_SUCCESS);
}

/*****************************************************************************
* stfp_getinfo - get device driver information
*/

static int
stfp_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
    struct stfp_unit *psu;
	
    DEBUG3(CE_CONT, "Entering stfp_getinfo:\n");
	
    switch (infocmd)
    {
        case DDI_INFO_DEVT2DEVINFO:
            if((psu = ddi_get_soft_state( stfpstatep,
                getminor((dev_t)arg ))) != NULL )
            {
                *result = psu->dip;
                return (DDI_SUCCESS);
            }
            break;

        case DDI_INFO_DEVT2INSTANCE:
            *result = (void *)getminor((dev_t)arg);
            return (DDI_SUCCESS);
            break;
		
        default:
            break;
    }
	
    return (DDI_FAILURE);
}

/*****************************************************************************
* stfp_open - gain access to the device
*/

static int
stfp_open(dev_t *devp, int falg, int otyp, cred_t *credp)
{
    struct stfp_unit *psu;

    DEBUG3(CE_CONT, "Entering stfp_open:\n");

    if(otyp != OTYP_CHR)    /* Support CHAR device only */
        return(EINVAL);     /* Return error if not */

    /* Return error if attach was not successful */
    if(ddi_get_soft_state(stfpstatep, getminor(*devp)) == NULL)
        return(ENXIO);

    /* Get device structure pointer */
    psu = ddi_get_soft_state(stfpstatep, getminor(*devp));

    return (0);
}

/*****************************************************************************
* stfp_close - relinquishe access to the device
*/

static int
stfp_close(dev_t dev, int flag, int otyp, cred_t *credp)
{
    struct stfp_unit *psu;

    DEBUG3(CE_CONT, "Entering stfp_close:\n");

    /* Get device structure pointer */
    psu = ddi_get_soft_state(stfpstatep, getminor(dev));

    return (0);
}

/*****************************************************************************
* stfp_read - read data from device
*/

static int
stfp_read(dev_t dev, struct uio *uiop, cred_t *credp)
{
    struct stfp_unit *psu;
    int n;  /* length of the GPS packet */
	
    DEBUG3(CE_CONT, "Entering stfp_read:\n");

    /* Get device structure pointer */
    psu = ddi_get_soft_state(stfpstatep, getminor(dev));

    /* Check availability of the GPS packet */
    if(STFP_DP_GETB(psu->pGps) == 0)
        return(EIO);

    /* Check length of the GPS packet */
    if((n = STFP_DP_GETB(psu->pGps) + 1) > STFP_MAX_READ)
        return(EIO);

    /* 64-bit */
    return (uiomove((caddr_t)psu->pGps, n, UIO_READ, uiop));
}

/*****************************************************************************
* stfp_write - write to device
*/

static int
stfp_write(dev_t dev, struct uio *uiop, cred_t *credp)
{
    struct stfp_unit *psu;
    int n;  /* length of the GPS packet */
    int retval = 0;
    uint32_t time;

    DEBUG3(CE_CONT, "Entering stfp_write:\n");

    /* Get device structure pointer */
    psu = ddi_get_soft_state(stfpstatep, getminor(dev));

    /* Check length of the written GPS packet */
    if((n = uiop->uio_iov->iov_len) > STFP_MAX_WRITE)
        return(EFBIG);

    retval = uiomove((caddr_t)psu->pIn, n, UIO_WRITE, uiop);
    if(retval == 0)
    {
        /* handle a responce if there is one */
        if ((STFP_DP_GETB(psu->pIn) == COM_MANREQGPSPACKET) &&
            (STFP_DP_GETB(psu->pIn+2) != 0))
        {
            /* do not sleep if INT_PACKET signal enabled */
            if(psu->intMaskSig & INT_PACKET)
                retval = sendCommand(psu);
            else
            {
                STFP_REG_ANDEQUAL(&psu->pRegs->mask, ~INT_PACKET);
                STFP_REG_PUT32(&psu->pRegs->intstat, INT_PACKET);
                STFP_REG_PUT32(&psu->pRegs->ack, ACK_PACKET);
                retval = sendCommand(psu);

                /* sleep if packet is not immediately available */
                mutex_enter(&psu->mutex);

                if(!(STFP_REG_GET32(&psu->pRegs->intstat) & INT_PACKET))
                {
                    STFP_REG_OREQUAL(&psu->pRegs->mask, INT_PACKET);
                    delay(drv_usectohz(1000000)); /* Wait 1s */	
                    STFP_REG_ANDEQUAL(&psu->pRegs->mask, ~INT_PACKET);
                }
                mutex_exit(&psu->mutex);
            }
        }
        else
            retval = sendCommand(psu);
    }
    return (retval);				
}

/*****************************************************************************
* stfp_ioctl - perform arbitrary operations
*/

static int
stfp_ioctl(dev_t dev, int cmd, intptr_t arg, int flag, cred_t *credp, int *rvalp)
{
    struct stfp_unit *psu;
    struct stfp_timeval Tv;
    struct stfp_time Tm;
    struct stfp_dec_tm dec;
    struct stfp_version Ver;
    struct periodic Per;
    struct tcgenoffset tcgoff;
    struct loctmoffset loctoff;
    struct leapseconds lpsec;

    struct getdata_t getdata;
    int get_data_len;
    u_char *getdataptr;
    int retval = 0;
    int32_t argument;
    uint32_t time;
    int i;
    void *pref;
    struct dp_rdwr_t dprdwr;
    struct dds_command_t ddscmd;
    struct tc_command_t tccmd;
    struct reg_rdwr_t regrdwr;

    DEBUG3(CE_CONT, "Entering stfp_ioctl:\n");

    psu = ddi_get_soft_state(stfpstatep, getminor(dev));

    switch(cmd)
    {
        case TIMEREQUEST:
            STFP_REG_PUT32(&psu->pRegs->timereq, 0);
            break;

        case EVENTREQUEST:
            STFP_REG_PUT32(&psu->pRegs->eventreq, 0);
            break;

        case RDTIME:
            rdTime(psu, psu->pRegs->time, psu->pYear, &Tm);
            if(ddi_copyout((caddr_t) &Tm, (caddr_t)arg,
                sizeof(struct stfp_time), flag) != 0)
                retval = EFAULT;
            break;

        case RDEVENT:
            rdTime(psu, psu->pRegs->event, psu->pYear, &Tm);
            if(ddi_copyout((caddr_t) &Tm, (caddr_t)arg,
                sizeof(struct stfp_time), flag) != 0)
                retval = EFAULT;
            break;

        case RDTIMEREQ:
            STFP_REG_PUT32(&psu->pRegs->timereq, 0);
            rdTime(psu, psu->pRegs->time, psu->pYear, &Tm);
            if(ddi_copyout((caddr_t) &Tm, (caddr_t)arg,
                sizeof(struct stfp_time), flag) != 0)
                retval = EFAULT;
            break;

        case RDEVENTREQ:
            STFP_REG_PUT32(&psu->pRegs->eventreq, 0);
            rdTime(psu, psu->pRegs->event, psu->pYear, &Tm);
            if(ddi_copyout((caddr_t) &Tm, (caddr_t)arg,
                sizeof(struct stfp_time), flag) != 0)
                retval = EFAULT;
            break;

        case RDTIMETV:
            rdTimeTv(psu, psu->pRegs->time, &Tv);
            if(ddi_copyout((caddr_t) &Tv, (caddr_t)arg,
                sizeof(struct stfp_timeval), flag) != 0)
                retval = EFAULT;
            break;

        case RDEVENTTV:
            rdTimeTv(psu, psu->pRegs->event, &Tv);
            if(ddi_copyout((caddr_t) &Tv, (caddr_t)arg,
                sizeof(struct stfp_timeval), flag) != 0)
                retval = EFAULT;
            break;

        case RDTIMETVREQ:
            STFP_REG_PUT32(&psu->pRegs->timereq, 0);
            rdTimeTv(psu, psu->pRegs->time, &Tv);
            if(ddi_copyout((caddr_t) &Tv, (caddr_t)arg,
                sizeof(struct stfp_timeval), flag) != 0)
                retval = EFAULT;
            break;

        case RDEVENTTVREQ:
            STFP_REG_PUT32(&psu->pRegs->eventreq, 0);
            rdTimeTv(psu, psu->pRegs->event, &Tv);
            if(ddi_copyout((caddr_t) &Tv, (caddr_t)arg,
                sizeof(struct stfp_timeval), flag) != 0)
                retval = EFAULT;
            break;

        case CAPUNLOCK:
            STFP_REG_PUT32(&psu->pRegs->unlock, 0);
            break;

        case RDINTSTAT:
            argument = STFP_REG_GET32(&psu->pRegs->intstat) & INT_BITMASK;
            if(ddi_copyout((caddr_t) &argument, (caddr_t)arg,
                sizeof(int), flag) != 0)
                retval = EFAULT;
            break;

        case CLRINTSTAT:
            STFP_REG_PUT32(&psu->pRegs->intstat, arg & ~psu->intMaskSig);
            break;
		
        case RDCONTROL:
            argument = STFP_REG_GET32(&psu->pRegs->control);
            if(ddi_copyout((caddr_t) &argument, (caddr_t)arg,
                sizeof(int), flag) != 0)
                retval = EFAULT;
            break;	

        case SETCONTROL:
            STFP_REG_PUT32(&psu->pRegs->control, arg);
            break;	

        case WRSTROBE:
            if(ddi_copyin((caddr_t)arg, (caddr_t) &Tm,
                sizeof(struct stfp_time), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            STFP_REG_ANDEQUAL(&psu->pRegs->control, ~CTRL_STREN);
            STFP_REG_PUT32(&psu->pRegs->mijstrb, ((Tm.tm.tm_hour << 16) + (Tm.tm.tm_min << 8) + Tm.tm.tm_sec) & 0xffffff);
            STFP_REG_PUT32(&psu->pRegs->minstrb, Tm.usec & 0xfffff);
            STFP_REG_OREQUAL(&psu->pRegs->control, CTRL_STREN);
            break;

        case WRSTROBETV:
            if(ddi_copyin((caddr_t)arg, (caddr_t) &Tv,
                sizeof(struct stfp_timeval), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            STFP_REG_ANDEQUAL(&psu->pRegs->control, ~CTRL_STREN);
            STFP_REG_PUT32(&psu->pRegs->mijstrb, Tv.tv.tv_sec);
            STFP_REG_PUT32(&psu->pRegs->minstrb, Tv.tv.tv_usec & 0xfffff);
            STFP_REG_OREQUAL(&psu->pRegs->control, CTRL_STREN);
            break;

        case SETTIME:   /* Binary Time Format */
            STFP_DP_PUTB(psu->pIn, COM_SETTIME);
            STFP_DP_PUTB(psu->pIn+1, (arg >> 24));
            STFP_DP_PUTB(psu->pIn+2, (arg >> 16));
            STFP_DP_PUTB(psu->pIn+3, (arg >> 8));
            STFP_DP_PUTB(psu->pIn+4, arg);
            retval = sendCommand(psu);
            delay(drv_usectohz(800000)); /* Wait 800ms */
            break;

        case SETDECTIME:    /* Decimal Time Format */
            if(ddi_copyin((caddr_t)arg, (caddr_t) &dec,
                sizeof(struct stfp_dec_tm), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            STFP_DP_PUTB(psu->pIn, COM_SETTIME);
            STFP_DP_PUTB(psu->pIn+1, dec.dec_year >> 8);
            STFP_DP_PUTB(psu->pIn+2, dec.dec_year);
            STFP_DP_PUTB(psu->pIn+3, dec.dec_day >> 8);
            STFP_DP_PUTB(psu->pIn+4, dec.dec_day);
            STFP_DP_PUTB(psu->pIn+5, dec.dec_hour);
            STFP_DP_PUTB(psu->pIn+6, dec.dec_min);
            STFP_DP_PUTB(psu->pIn+7, dec.dec_sec);
            retval = sendCommand(psu);
            delay(drv_usectohz(800000)); /* Wait 800ms */
            break;

        case SETTIMETV: /* Get time from tv structure */
            if(ddi_copyin((caddr_t)arg, (caddr_t) &Tv,
                sizeof(struct stfp_timeval), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            STFP_DP_PUTB(psu->pIn, COM_SETTIME);
            STFP_DP_PUTB(psu->pIn+1, (Tv.tv.tv_sec >> 24));
            STFP_DP_PUTB(psu->pIn+2, (Tv.tv.tv_sec >> 16));
            STFP_DP_PUTB(psu->pIn+3, (Tv.tv.tv_sec >> 8));
            STFP_DP_PUTB(psu->pIn+4, Tv.tv.tv_sec);
            retval = sendCommand(psu);
            delay(drv_usectohz(800000)); /* Wait 800ms */
            break;

        case SETYEAR:
            STFP_DP_PUTB(psu->pIn+0, COM_SETYEAR);
            STFP_DP_PUTB(psu->pIn+1, arg >> 8);
            STFP_DP_PUTB(psu->pIn+2, arg);
            retval = sendCommand(psu);
            delay(drv_usectohz(800000)); /* Wait 800ms */
            break;

        case SETINTSIGNAL:
            mutex_enter(&psu->mutex);

            psu->intMaskSig = arg & INT_BITMASK;
            if (psu->intMaskSig)
            {
                if (!psu->pref)
                    psu->pref = proc_ref();
            }
            else
            {
                if (psu->pref)
                {
                    pref = psu->pref;
                    psu->pref = 0;
                    proc_unref(pref);
                }
            }

            STFP_REG_PUT32(&psu->pRegs->intstat, psu->intMaskSig);
            if(psu->intMaskSig & INT_PACKET)
                STFP_REG_PUT32(&psu->pRegs->ack, ACK_PACKET);

            STFP_REG_PUT32(&psu->pRegs->mask, (STFP_REG_GET32(&psu->pRegs->mask) & INT_PACKET) | psu->intMaskSig);

            mutex_exit(&psu->mutex);
            break;

        case RDINTSIGNAL:
            mutex_enter(&psu->mutex);
            argument = psu->intServiced & INT_BITMASK;
            psu->intServiced = 0;
            mutex_exit(&psu->mutex);
            if(ddi_copyout((caddr_t) &argument, (caddr_t)arg,
                sizeof(int), flag) != 0)
                retval = EFAULT;
            break;

        case REQGPSPACKET:
            STFP_DP_PUTB(psu->pIn, COM_REQGPSPACKET);
            STFP_DP_PUTB(psu->pIn+1, arg);

            /* do not sleep if INT_PACKET signal enabled */
            if(psu->intMaskSig & INT_PACKET)
                retval = sendCommand(psu);
            else
            {
                STFP_REG_ANDEQUAL(&psu->pRegs->mask, ~INT_PACKET);
                STFP_REG_PUT32(&psu->pRegs->intstat, INT_PACKET);
                STFP_REG_PUT32(&psu->pRegs->ack, ACK_PACKET);
                retval = sendCommand(psu);

                /* sleep if packet is not immediately available */
                mutex_enter(&psu->mutex);

                if(!(STFP_REG_GET32(&psu->pRegs->intstat) & INT_PACKET))
                {
                    STFP_REG_OREQUAL(&psu->pRegs->mask, INT_PACKET);
                    drv_getparm(LBOLT, &time);  /* ticks since reboot */
                    time += drv_usectohz(5e6);  /* add 5 seconds */

                    if(cv_timedwait(&psu->cv, &psu->mutex, time) == 0)
                    {
                        /* if timeout */
                        retval = EFAULT;
                        mutex_exit(&psu->mutex);
                        break;
                    }
                    STFP_REG_ANDEQUAL(&psu->pRegs->mask, ~INT_PACKET);
                }
                mutex_exit(&psu->mutex);
            }
            break;

        case CONTROLTIMEBASE:
            switch (arg)
            {
                case CLOCK_INTERNAL:
                    STFP_DP_PUTB(psu->pIn, COM_SELCLOCK);
                    STFP_DP_PUTB(psu->pIn+1, SELCLOCK_INT);
                    retval = sendCommand(psu);
                    break;
                case CLOCK_EXTERNAL:
                    STFP_DP_PUTB(psu->pIn, COM_SELCLOCK);
                    STFP_DP_PUTB(psu->pIn+1, SELCLOCK_EXT);
                    retval = sendCommand(psu);
                    break;
                case JAMSYNC_ENABLE:
                    STFP_DP_PUTB(psu->pIn, COM_JAMSYNCENABLE);
                    STFP_DP_PUTB(psu->pIn+1, JAMSYNC_ENABLE);
                    retval = sendCommand(psu);
                    break;
                case JAMSYNC_DISABLE:
                    STFP_DP_PUTB(psu->pIn, COM_JAMSYNCENABLE);
                    STFP_DP_PUTB(psu->pIn+1, JAMSYNC_DISABLE);
                    retval = sendCommand(psu);
                    break;
                case DISCIPLINE_ENABLE:
                    STFP_DP_PUTB(psu->pIn, COM_DISCENABLE);
                    STFP_DP_PUTB(psu->pIn+1, DISC_EN);
                    retval = sendCommand(psu);
                    break;
                case DISCIPLINE_DISABLE:
                    STFP_DP_PUTB(psu->pIn, COM_DISCENABLE);
                    STFP_DP_PUTB(psu->pIn+1, DISC_DIS);
                    retval = sendCommand(psu);
                    break;
                case FORCE_JAMSYNC:
                    STFP_DP_PUTB(psu->pIn, COM_FORCEJAMSYNC);
                    retval = sendCommand(psu);
                    break;
                case CLEAR_USEC:
                    break;
                default:
                    retval = EINVAL;
                    break;
            }
            break;

        case SETDAC:
            STFP_DP_PUTB(psu->pIn, COM_SETDAC);
            STFP_DP_PUTB(psu->pIn+1, arg >> 8);
            STFP_DP_PUTB(psu->pIn+2, arg);
            retval = sendCommand(psu);
            break;

        case RDDAC:
            STFP_DP_PUTB(psu->pIn, COM_GETDATA);
            STFP_DP_PUTB(psu->pIn+1, GETDATA_DAC);
            if(retval = sendCommand(psu) != 0)
                break;
            argument = (STFP_DP_GETB(psu->pOut+1) << 8) & 0xff00;
            argument |= (STFP_DP_GETB(psu->pOut+2) & 0x00ff);
            if(ddi_copyout((caddr_t) &argument, (caddr_t)arg,
                sizeof(int), flag) != 0)
                retval = EFAULT;
            break;

        case SETTIMINGOFFSET:
            STFP_DP_PUTB(psu->pIn, COM_SETTIMINGOFFSET);
            STFP_DP_PUTB(psu->pIn+1, arg >> 24);
            STFP_DP_PUTB(psu->pIn+2, arg >> 16);
            STFP_DP_PUTB(psu->pIn+3, arg >> 8);
            STFP_DP_PUTB(psu->pIn+4, arg);
            retval = sendCommand(psu);
            break;

        case SELTIMINGMODE:
            STFP_DP_PUTB(psu->pIn, COM_SETTIMINGMODE);
            STFP_DP_PUTB(psu->pIn+1, arg);
            retval = sendCommand(psu);
            delay(drv_usectohz(1000000)); /* wait 1sec to settle */
            break;

        case SELTIMEFORMAT:
            STFP_DP_PUTB(psu->pIn, COM_SETTIMEFORMAT);
            STFP_DP_PUTB(psu->pIn+1, arg);
            retval = sendCommand(psu);
            delay(drv_usectohz(1000000)); /* wait 1sec to settle */
            break;

        case SELTCFORMAT:
            STFP_DP_PUTB(psu->pIn, COM_SETTCFORMAT);
            STFP_DP_PUTB(psu->pIn+1, arg);
            retval = sendCommand(psu);
            delay(drv_usectohz(1000000)); /* wait 1sec to settle */
            break;

        case SELTCMOD:
            STFP_DP_PUTB(psu->pIn, COM_SETTCMOD);
            STFP_DP_PUTB(psu->pIn+1, arg);
            retval = sendCommand(psu);
            delay(drv_usectohz(1000000)); /* wait 1sec to settle */
            break;

        case SETPERIODIC:
            if(ddi_copyin((caddr_t)arg, (caddr_t) &Per,
                sizeof(struct periodic), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            STFP_DP_PUTB(psu->pIn, COM_SETPERIODIC);
            STFP_DP_PUTB(psu->pIn+1, Per.sync1pps);
            STFP_DP_PUTB(psu->pIn+2, Per.n1 >> 8);
            STFP_DP_PUTB(psu->pIn+3, Per.n1);
            STFP_DP_PUTB(psu->pIn+4, Per.n2 >> 8);
            STFP_DP_PUTB(psu->pIn+5, Per.n2);

            DEBUG3(CE_CONT, "SETPERIODIC: pIn = 0x%x, 0x%x,0x%x,0x%x,0x%x,0x%x\n",
                STFP_DP_GETB(psu->pIn),
                STFP_DP_GETB(psu->pIn+1),
                STFP_DP_GETB(psu->pIn+2),
                STFP_DP_GETB(psu->pIn+3),
                STFP_DP_GETB(psu->pIn+4),
                STFP_DP_GETB(psu->pIn+5)
					   );
            retval = sendCommand(psu);
            break;

        case CONTROLEVENT:
            switch (arg)
            {
                case EVENT_RISING:
                    STFP_REG_ANDEQUAL(&psu->pRegs->control, ~CTRL_EVSENSE);
                    break;
                case EVENT_FALLING:
                    STFP_REG_OREQUAL(&psu->pRegs->control, CTRL_EVSENSE);
                    break;
                case CAPLOCK_ENABLE:
                    STFP_REG_OREQUAL(&psu->pRegs->control, CTRL_LOCKEN);
                    break;
                case CAPLOCK_DISABLE:
                    STFP_REG_ANDEQUAL(&psu->pRegs->control, ~CTRL_LOCKEN);
                    break;
                case EVCAP_ENABLE:
                    STFP_REG_OREQUAL(&psu->pRegs->control, CTRL_EVENTEN);
                    break;
                case EVCAP_DISABLE:
                    STFP_REG_ANDEQUAL(&psu->pRegs->control, ~CTRL_EVENTEN);
                    break;
                case EVCAP_PERIODIC:
                    STFP_REG_OREQUAL(&psu->pRegs->control, CTRL_EVSOURCE);
                    break;
                case EVCAP_EVENT:
                    STFP_REG_ANDEQUAL(&psu->pRegs->control, ~CTRL_EVSOURCE);
                    break;
                default:
                    retval = EINVAL;
                    break;
            }
            break;

        case CONTROLSTROBE:
            switch (arg)
            {
                case STROBE_SECUS:
                    STFP_REG_ANDEQUAL(&psu->pRegs->control, ~CTRL_STRMODE);
                    break;
                case STROBE_USONLY:
                    STFP_REG_OREQUAL(&psu->pRegs->control, CTRL_STRMODE);
                    break;
                case STROBE_ENABLE:
                    STFP_REG_OREQUAL(&psu->pRegs->control, CTRL_STREN);
                    break;
                case STROBE_DISABLE:
                    STFP_REG_ANDEQUAL(&psu->pRegs->control, ~CTRL_STREN);
                    break;
                default:
                    retval = EINVAL;
                    break;
            }
            break;

        case SELFREQUENCYOUT:
            switch (arg)
            {
                case FREQ_10MHZ:
                    STFP_REG_PUT32(&psu->pRegs->control, (STFP_REG_GET32(&psu->pRegs->control) & ~CTRL_FMASK) | CTRL_F10);
                    break;
                case FREQ_5MHZ:
                    STFP_REG_PUT32(&psu->pRegs->control, (STFP_REG_GET32(&psu->pRegs->control) & ~CTRL_FMASK) | CTRL_F5);
                    break;
                case FREQ_1MHZ:
                    STFP_REG_PUT32(&psu->pRegs->control, (STFP_REG_GET32(&psu->pRegs->control) & ~CTRL_FMASK) | CTRL_F1);
                    break;
                default:
                    retval = EINVAL;
                    break;
            }
            break;

        case SOFTWARERESET:
            STFP_DP_PUTB(psu->pIn, COM_SOFTRESET);
            STFP_DP_PUTB(psu->pIn+1, SW_RESET);
            retval = sendCommand(psu);
            delay(drv_usectohz(1000000)); /* wait 1sec to settle */
            break;

        case SETTCOUTFMT:
            STFP_DP_PUTB(psu->pIn, COM_SETTCOUTFMT);
            STFP_DP_PUTB(psu->pIn+1, arg);
            retval = sendCommand(psu);
            break;

        case SETGENTMOFFSET:
            if(ddi_copyin((caddr_t)arg, (caddr_t) &tcgoff,
                sizeof(struct tcgenoffset), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            STFP_DP_PUTB(psu->pIn, COM_SETGENTMOFF);
            STFP_DP_PUTB(psu->pIn+1, tcgoff.genoff >> 8);
            STFP_DP_PUTB(psu->pIn+2, tcgoff.genoff);
            STFP_DP_PUTB(psu->pIn+3, tcgoff.genflg);
            retval = sendCommand(psu); 
            break;

        case SETLOCTMOFFSET:
            if(ddi_copyin((caddr_t)arg, (caddr_t) &loctoff,
                sizeof(struct loctmoffset), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            STFP_DP_PUTB(psu->pIn, COM_SETLOCTMOFF);
            STFP_DP_PUTB(psu->pIn+1, loctoff.locoff >> 8);
            STFP_DP_PUTB(psu->pIn+2, loctoff.locoff);
            STFP_DP_PUTB(psu->pIn+3, loctoff.locflg);
            retval = sendCommand(psu); 
            break;
			
        case SETLEAPSECEVENT:
            if(ddi_copyin((caddr_t)arg, (caddr_t) &lpsec,
                sizeof(struct leapseconds), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            STFP_DP_PUTB(psu->pIn, COM_SETLEAPSEC);
            STFP_DP_PUTB(psu->pIn+1, lpsec.ls_flag);
            STFP_DP_PUTB(psu->pIn+2, lpsec.leap_tm >> 24);
            STFP_DP_PUTB(psu->pIn+3, lpsec.leap_tm >> 16);
            STFP_DP_PUTB(psu->pIn+4, lpsec.leap_tm >> 8);
            STFP_DP_PUTB(psu->pIn+5, lpsec.leap_tm);
            retval = sendCommand(psu);
            break;

        case SETDISGAIN:
            STFP_DP_PUTB(psu->pIn, COM_SETDISCGAIN);
            STFP_DP_PUTB(psu->pIn+1, arg >> 8);
            STFP_DP_PUTB(psu->pIn+2, arg);
            retval = sendCommand(psu);
            break;

        case SYNCRTC:
            STFP_DP_PUTB(psu->pIn, COM_SYNCRTC);
            retval = sendCommand(psu);
            break;

        case DISCBATT:
            STFP_DP_PUTB(psu->pIn, COM_DISCBATTRTC);
            retval = sendCommand(psu);
            break;

        case SETCLKVAL:
            STFP_DP_PUTB(psu->pIn, COM_SETCLKVAL);
            STFP_DP_PUTB(psu->pIn+1, arg >> 24);
            STFP_DP_PUTB(psu->pIn+2, arg >> 16);
            STFP_DP_PUTB(psu->pIn+3, arg >> 8);
            STFP_DP_PUTB(psu->pIn+4, arg);
            retval = sendCommand(psu);
            break;

        case SETGPSTMFMT:
            STFP_DP_PUTB(psu->pIn, COM_GPSTIMEFORMAT);
            STFP_DP_PUTB(psu->pIn+1, arg);
            retval = sendCommand(psu);
            break;

        case SETGPSMDFLG:
            STFP_DP_PUTB(psu->pIn, COM_SETGPSMDFLG);
            STFP_DP_PUTB(psu->pIn+1, arg);
            retval = sendCommand(psu);
            break;

        case SETLOCTMFLG:
            STFP_DP_PUTB(psu->pIn, COM_SETLOCTMFLG);
            STFP_DP_PUTB(psu->pIn+1, arg);
            retval = sendCommand(psu);
            break;

        case SETYRINCFLG:
            STFP_DP_PUTB(psu->pIn, COM_SETYRINCFLG);
            STFP_DP_PUTB(psu->pIn+1, arg);
            retval = sendCommand(psu);
            break;

        case RDFIRMVERSION:
            STFP_DP_PUTB(psu->pIn, COM_GETDATA);
            STFP_DP_PUTB(psu->pIn+1, GETDATA_VERSION);
            if(retval = sendCommand(psu) != 0)
                break;
            Ver.vmajor = STFP_DP_GETB(psu->pOut+1);
            Ver.vminor = STFP_DP_GETB(psu->pOut+2);
            Ver.month = STFP_DP_GETB(psu->pOut+3);
            Ver.day = STFP_DP_GETB(psu->pOut+4);
            Ver.year = STFP_DP_GETB(psu->pOut+5) << 8 | STFP_DP_GETB(psu->pOut+6);
            if(ddi_copyout((caddr_t) &Ver, (caddr_t)arg,
                sizeof(struct stfp_version), flag) != 0)
                retval = EFAULT;
            break;

        case GETDATA:
            if(ddi_copyin((caddr_t)arg, (caddr_t) &getdata,
                sizeof(struct getdata_t), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            STFP_DP_PUTB(psu->pIn, COM_GETDATA);
            STFP_DP_PUTB(psu->pIn+1, getdata.arg);
            if(retval = sendCommand(psu) != 0)
                break;

            /* get the number of bytes to be transferred */
            switch(getdata.arg)
            {
                case GETDATA_MODE:
                    get_data_len=sizeof(getdata.data.tmode);
                    break;
                case GETDATA_TFORMAT:
                    get_data_len=sizeof(getdata.data.tformat);
                    break;
                case GETDATA_YEAR:
                    get_data_len=sizeof(getdata.data.year);
                    break;
                case GETDATA_PER:
                    get_data_len=sizeof(getdata.data.period);
                    break;
                case GETDATA_TCFMT:
                    get_data_len=sizeof(getdata.data.tcformat);
                    break;
                case GETDATA_TCMOD:
                    get_data_len=sizeof(getdata.data.tcmod);
                    break;
                case GETDATA_TOFF:
                    get_data_len=sizeof(getdata.data.offset);
                    break;
                case GETDATA_UTCINFO:
                    get_data_len=sizeof(getdata.data.utc);
                    break;
                case GETDATA_TCOUTFMT:
                    get_data_len=sizeof(getdata.data.tcoutfmt);
                    break;
                case GETDATA_TCGENOFF:
                    get_data_len=sizeof(getdata.data.genoff);
                    break;
                case GETDATA_LOCTMOFF:
                    get_data_len=sizeof(getdata.data.locoff);
                    break;
                case GETDATA_LEAPSEC:
                    get_data_len=sizeof(getdata.data.leapsec);
                    break;
                case GETDATA_FWVER:
                    get_data_len=sizeof(getdata.data.version);
                    break;
                case GETDATA_CLKSRC:
                    get_data_len=sizeof(getdata.data.clksrc);
                    break;
                case GETDATA_JAMSC:
                    get_data_len=sizeof(getdata.data.jsctl);
                    break;
                case GETDATA_OSCCTL:
                    get_data_len=sizeof(getdata.data.ocsctl);
                    break;
                case GETDATA_DAVAL:
                    get_data_len=sizeof(getdata.data.davalue);
                    break;
                case GETDATA_BATTSTAT:
                    get_data_len=sizeof(getdata.data.battstat);
                    break;
                case GETDATA_CLKVAL:
                    get_data_len=sizeof(getdata.data.clkval);
                    break;
                case GETDATA_DTFW:
                    get_data_len=sizeof(getdata.data.dtfw);
                    break;
                case GETDATA_ASSEMB:
                    get_data_len=sizeof(getdata.data.assemb);
                    break;
                case GETDATA_HWFAB:
                    get_data_len=sizeof(getdata.data.hwfab);
                    break;
                case GETDATA_TFPMODEL:
                    get_data_len=sizeof(getdata.data.tfpm);
                    break;
                case GETDATA_SERIAL:
                    get_data_len=sizeof(getdata.data.serialnum);
                    break;
                default:
                    get_data_len=0;
                    retval=EFAULT;
            }
            /* Skip over response byte. Only return data */
            getdataptr=(u_char *)&(getdata.data.tmode);

            for(i=0; i<get_data_len; i++)
                getdataptr[i] = STFP_DP_GETB(psu->pOut+i+1);

            if(ddi_copyout((caddr_t) &getdata, (caddr_t)arg, 
                sizeof(struct getdata_t), flag) != 0)
            {
                retval = EFAULT;
            }
            break;

        case REQREVID:
            argument = psu->revID;
            if (ddi_copyout((caddr_t) &argument, (caddr_t)arg,
                sizeof(int), flag) != 0)
                retval = EFAULT;
            break;

        case RDDPOFFSET:
            if (ddi_copyin((caddr_t)arg, (caddr_t) &dprdwr,
                sizeof(struct dp_rdwr_t), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            dprdwr.data = STFP_DP_GETB(psu->pDpr+dprdwr.offset);
            DEBUG3(CE_CONT, "ioctl (RDDPOFFSET): offset=0x%x, data=0x%x\n",
                    (uint32_t)dprdwr.offset, (uint32_t)dprdwr.data);
            if (ddi_copyout((caddr_t) &dprdwr, (caddr_t)arg,
                sizeof(struct dp_rdwr_t), flag) != 0)
                retval = EFAULT;
            break;

        case WRDPOFFSET:
            if (ddi_copyin((caddr_t)arg, (caddr_t) &dprdwr,
                sizeof(struct dp_rdwr_t), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            DEBUG3(CE_CONT, "ioctl (WRDPOFFSET): offset=0x%x, data=0x%x\n",
                    (uint32_t)dprdwr.offset, (uint32_t)dprdwr.data);
            STFP_DP_PUTB(psu->pDpr+dprdwr.offset, dprdwr.data);
            break;

        case SENDCOMMAND:
            retval = sendCommand(psu);
            break;

        case DDSCOMMAND:
            if (psu->revID >= BC_PCI_V2_REV_ID_START &&
                psu->revID <= BC_PCI_V2_REV_ID_END)
            {
                if (ddi_copyin((caddr_t)arg, (caddr_t) &ddscmd,
                    sizeof(struct dds_command_t), flag) != 0)
                {
                    retval = EFAULT;
                    break;
                }
                DEBUG3(CE_CONT, "ioctl (DDSCOMMAND): id=0x%x, len=%d, data=0x%x, 0x%x, 0x%x, 0x%x\n",
                        (uint32_t)ddscmd.cmd_id, (uint32_t)ddscmd.cmd_len,
                        (uint32_t)ddscmd.data.cmd_byte[0],
                        (uint32_t)ddscmd.data.cmd_byte[1],
                        (uint32_t)ddscmd.data.cmd_byte[2],
                        (uint32_t)ddscmd.data.cmd_byte[3]);
                if (ddscmd.cmd_id == COM_SET_DDS_PERIOD)
                {
                    DEBUG3(CE_CONT, "    DDS_PERIOD: %d\n",
                            ddscmd.data.cmd_number);
                }
                if (ddscmd.cmd_id == COM_SET_DDS_TUNEWORD)
                {
                    DEBUG3(CE_CONT, "    DDS_TUNEWORD: %d\n",
                            ddscmd.data.cmd_number);
                }
                STFP_DP_PUTB(psu->pIn, ddscmd.cmd_id);
                for (i = 0; i < ddscmd.cmd_len; i++)
                    STFP_DP_PUTB(psu->pIn+i+1, ddscmd.data.cmd_byte[i]);
                retval = sendCommand(psu);
            }
            else
            {
                DEBUG1(CE_WARN, "DDS commands available on V2 hardware only\n");
                retval = ENOTTY;
            }
            break;

        case TCFORMATEX:
            if (ddi_copyin((caddr_t)arg, (caddr_t) &tccmd,
                sizeof(struct tc_command_t), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            DEBUG3(CE_CONT, "ioctl (TCFORMATEX): id=0x%x, type=%c, subtype=%c\n",
                    (uint32_t)tccmd.cmd_id, tccmd.type, tccmd.subtype);
            STFP_DP_PUTB(psu->pIn, tccmd.cmd_id);
            STFP_DP_PUTB(psu->pIn+1, tccmd.type);
            STFP_DP_PUTB(psu->pIn+2, tccmd.subtype);
            retval = sendCommand(psu);
            break;

        case RDREGOFFSET:
            if (ddi_copyin((caddr_t)arg, (caddr_t) &regrdwr,
                sizeof(struct reg_rdwr_t), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            regrdwr.data = STFP_REG_GET32((uint32_t *)((uchar_t *)psu->pRegs+regrdwr.offset));
            DEBUG3(CE_CONT, "ioctl (RDREGOFFSET): offset=0x%x, data=0x%x\n",
                    (uint32_t)regrdwr.offset, regrdwr.data);
            if (ddi_copyout((caddr_t) &regrdwr, (caddr_t)arg,
                sizeof(struct reg_rdwr_t), flag) != 0)
                retval = EFAULT;
            break;

        case WRREGOFFSET:
            if (ddi_copyin((caddr_t)arg, (caddr_t) &regrdwr,
                sizeof(struct reg_rdwr_t), flag) != 0)
            {
                retval = EFAULT;
                break;
            }
            DEBUG3(CE_CONT, "ioctl (WRREGOFFSET): offset=0x%x, data=0x%x\n",
                    (uint32_t)regrdwr.offset, regrdwr.data);
            STFP_REG_PUT32((uint32_t *)((uchar_t *)psu->pRegs+regrdwr.offset), regrdwr.data);
            break;

        case RDEVENT2TV:
            rdTimeTv(psu, psu->pRegs->event2, &Tv);
            if(ddi_copyout((caddr_t) &Tv, (caddr_t)arg,
                sizeof(struct stfp_timeval), flag) != 0)
                retval = EFAULT;
            break;

        case RDEVENT3TV:
            rdTimeTv(psu, psu->pRegs->event3, &Tv);
            if(ddi_copyout((caddr_t) &Tv, (caddr_t)arg,
                sizeof(struct stfp_timeval), flag) != 0)
                retval = EFAULT;
            break;

        default:
            retval = ENOTTY;
            break;
    }
    return (retval);
}

/*------------------------ INTERRUPT HANDLER --------------------------------*/

/*****************************************************************************
* stfp_intr - low level interrupt handler
*/
static u_int
stfp_intr(caddr_t arg)
{
    struct stfp_unit *psu = (struct stfp_unit *)arg;
    volatile int32_t pending;

    DEBUG3(CE_CONT, "Enter stfp_intr:");

    mutex_enter(&psu->mutex);
    pending = STFP_INTR_PENDING;
    DEBUG3(CE_CONT, " 0x%X\n", pending);

    /* claim or reject interrupt check */
    if(pending == 0)
    {
        mutex_exit(&psu->mutex);
        DEBUG3(CE_CONT, "Interrupt UNCLAIMED!!!\n");
        return(DDI_INTR_UNCLAIMED);
    }

    psu->intServiced |= pending & psu->intMaskSig;
    STFP_REG_PUT32(&psu->pRegs->intstat, pending);

    if(pending & INT_PACKET)
        STFP_REG_PUT32(&psu->pRegs->ack, ACK_PACKET);

    if (psu->pref != 0 && pending & psu->intMaskSig)
    {
        if (proc_signal(psu->pref, SIGUSR1) == -1)
        {
            proc_unref(psu->pref);
            psu->pref = 0;
        }
    }

    if(pending & ~psu->intMaskSig & INT_PACKET)
        cv_broadcast(&psu->cv);

    mutex_exit(&psu->mutex);

    return(DDI_INTR_CLAIMED);
}

/*------------------------ USER FUNCTIONS -----------------------------------*/

/*****************************************************************************
* rdTmTv - read time from the board in binary seconds format
*/
void rdTimeTv(struct stfp_unit *psu, uint32_t *pTmReg, struct stfp_timeval *pTmT)
{
    pTmT->tv.tv_sec = STFP_REG_GET32(pTmReg+1);
    pTmT->tv.tv_usec = STFP_REG_GET32(pTmReg) & 0xfffff;
    pTmT->hnsec = (STFP_REG_GET32(pTmReg) >> 20) & 0xf;
    pTmT->status = (STFP_REG_GET32(pTmReg) >> 24) & 0xf;
}

/*****************************************************************************
* rdTime - read time from the board in days through seconds format
*/
void rdTime(struct stfp_unit *psu, uint32_t *pTmReg, u_char *pY, struct stfp_time *pTmS)
{
    pTmS->tm.tm_year = (STFP_DP_GETB(pY) << 8) + 
            (STFP_DP_GETB(pY+1) - TM_YEAR_BASE);
    pTmS->tm.tm_yday = ((STFP_REG_GET32(pTmReg) >> 20)&0xf00) +
            ((STFP_REG_GET32(pTmReg+1) >> 24)&0xff) - 1;
    pTmS->tm.tm_hour = (STFP_REG_GET32(pTmReg+1) >> 16) & 0xff;
    pTmS->tm.tm_min = (STFP_REG_GET32(pTmReg+1) >> 8) & 0xff;
    pTmS->tm.tm_sec = STFP_REG_GET32(pTmReg+1) & 0xff;
    pTmS->usec = STFP_REG_GET32(pTmReg) & 0xfffff;
    pTmS->hnsec = (STFP_REG_GET32(pTmReg) >> 20) & 0xf;
    pTmS->status = (STFP_REG_GET32(pTmReg) >> 24) & 0xf;
}

/*****************************************************************************
* sendCommand - send command to the board to take action
*/
int sendCommand(struct stfp_unit *psu)
{
    int i;

    STFP_REG_PUT32(&psu->pRegs->ack, ACK_COMMAND);  /* clear ack command done bit */
    STFP_REG_PUT32(&psu->pRegs->ack, ACK_ACTION);   /* force bc635PCI to take action */

    for (i=0;((STFP_REG_GET32(&psu->pRegs->ack) & ACK_COMMAND) == 0) && i<20; i++)
        delay(drv_usectohz(1<<i)); /* Binary Backoff */

    if(i >= 20) /* check if timeout occured */
    {
        DEBUG3(CE_CONT, "sendCommand() timeout\n");
        return (EIO);
    }

/*  if (STFP_DP_GETB(psu->pIn) == 0)
    return (EINVAL); */

    return (0);
}
